from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import os

class Cluster_statistics:
    """
    Class to that holds statistics on the clusters.
    """

    def __init__(self, num_clusters: int, silhouette_score: float, centroids: list, radius: list[float], labels_percentages: list[float]):
        """
        Initialize the Cluster_statistics object with the number of clusters, inertia, and silhouette score.
        """
        self.num_clusters = num_clusters
        self.silhouette_score = silhouette_score
        self.centroids = centroids
        self.radius = radius
        self.labels_percentages = labels_percentages

    def __repr__(self):
        """
        String representation of the Cluster_statistics object.
        """

        radius_str = ', '.join(f'{r:.4f}' for r in self.radius)
        centroids_str = ', '.join(f'[{", ".join(f"{c:.4f}" for c in centroid)}]' for centroid in self.centroids)

        return (f"Cluster_statistics(num_clusters={self.num_clusters},\n"
                f"silhouette_score={self.silhouette_score:.4f},\n"
                f"centroids={centroids_str},\n"
                f"radius={radius_str},\n"
                f"labels_percentages={self.labels_percentages})")
    
    def get_json(self):
        """
        Convert the Cluster_statistics object to a JSON serializable dictionary.
        """
        centroid_list = [centroid.tolist() for centroid in self.centroids]
        radius_list = [radius for radius in self.radius]
        labels_percentages_list = [label_percentage for label_percentage in self.labels_percentages]

        return {
            "num_clusters": self.num_clusters,
            "silhouette_score": self.silhouette_score,
            "centroids": centroid_list,
            "radius": radius_list,
            "labels_percentages": labels_percentages_list
        }
    
    def __eq__(self, value):
        if not isinstance(value, Cluster_statistics):
            return NotImplemented
        
        if self.num_clusters != value.num_clusters:
            return False
        
        if self.silhouette_score != value.silhouette_score:
            return False
        
        value = ClusterChanges.reorder_changes(self, value)
        
        result = True
        i = 0
        while result and i < len(self.centroids):
            if not np.allclose(self.centroids[i], value.centroids[i]):
                result = False
            i += 1
        
        if not result:
            return False
        
        if not np.allclose(self.radius, value.radius):
            return False

        return np.allclose(self.labels_percentages, value.labels_percentages)

class ClusterChanges:
    """
    A class to represent the changes in clustering between two datasets.
    """

    def __init__(self, original: Cluster_statistics, new_data: Cluster_statistics, delta: float = 0.1):
        """
        Initializes the ClusterChanges with statistics from two clustering results.
        """
        self.original = original
        self.new_data = new_data
        self.delta = delta

        self.changed = dict()
        self.unchanged = dict()

        if original.num_clusters != new_data.num_clusters:
            self.changed['num_clusters'] = 100 * abs(original.num_clusters - new_data.num_clusters) / original.num_clusters
        else:
            self.unchanged['num_clusters'] = 100 * abs(original.num_clusters - new_data.num_clusters) / original.num_clusters

        metrics = [
            ("silhouette_score", original.silhouette_score, new_data.silhouette_score, delta * original.silhouette_score),
        ]

        if original.num_clusters == new_data.num_clusters:
            new_data = ClusterChanges.reorder_changes(original, new_data)

            # Compare centroids and radius
            for i, (orig_centroid, new_centroid, orig_radius) in enumerate(zip(original.centroids, new_data.centroids, original.radius)):
                metrics.append(
                    (f'centroid_{i}', np.linalg.norm(orig_centroid), np.linalg.norm(new_centroid), delta * orig_radius)
                )

            for i, (orig_radius, new_radius) in enumerate(zip(original.radius, new_data.radius)):
                metrics.append(
                    (f'radius_{i}', orig_radius, new_radius, delta * orig_radius)
                )

            for i, (orig_label_percentage, new_label_percentage) in enumerate(zip(original.labels_percentages, new_data.labels_percentages)):
                metrics.append(
                    (f'label_percentage_{i}', orig_label_percentage, new_label_percentage, delta * orig_label_percentage)
                )

        for name, orig_val, new_val, threshold in metrics:
            if type(orig_val) is np.ndarray and type(new_val) is np.ndarray:
                diff = np.linalg.norm(orig_val - new_val)
            else:
                diff = abs(orig_val - new_val)
            # Avoid division by zero
            if threshold == 0:
                percentage_diff = 0
            else:
                percentage_diff = diff / abs(threshold)
                if diff > abs(threshold):
                    self.changed[name] = int(percentage_diff * 100)
                else:
                    self.unchanged[name] = int(percentage_diff * 100)

    def __repr__(self):
        """
        Returns a string representation of the ClusterChanges.
        """
        change_str = ', '.join([f"{k}: {v}%" for k, v in self.changed.items()])
        unchanged_str = ', '.join([f"{k}: {v}%" for k, v in self.unchanged.items()])

        return (f"Change: {change_str}, Unchanged: {unchanged_str}, delta: {self.delta}")
    
    def reorder_changes(original, new_data):
        """
        Reorder the changes to match the order of another ClusterChanges object.
        """
        order = [0]*original.num_clusters
        for i in range(original.num_clusters):
            min_distance = float('inf')
            for j in range(new_data.num_clusters):
                distance = np.linalg.norm(original.centroids[i] - new_data.centroids[j])
                if distance < min_distance:
                    min_distance = distance
                    order[i] = j     
        
        # Order centroids and related statistics by closest new centroid positions
        new_data.centroids = [new_data.centroids[i] for i in order]
        new_data.radius = [new_data.radius[i] for i in order]
        new_data.labels_percentages = [new_data.labels_percentages[i] for i in order]

        return new_data

def _calculate_radius(X, kmeans):
    """
    Calculate the radius of each cluster based on the distance of points to their respective centroids.
    """

    X = np.array(X, dtype=float)
    radius = [0]*len(kmeans.cluster_centers_)
    for data_point, label in zip(X, kmeans.labels_):
        radius[label] = max(radius[label], np.linalg.norm(data_point-kmeans.cluster_centers_))

    return radius

def plot_clusters(X, kmeans, best_cluster, path: str = None):
    """
    Plot the clusters and their centroids.
    """
    if len(X.columns) == 2:
        plt.figure(figsize=(8, 4))
        plt.scatter(X.iloc[:, 0], X.iloc[:, 1], c=kmeans.labels_, cmap='viridis', marker='o')
        plt.scatter(kmeans.cluster_centers_[:, 0], kmeans.cluster_centers_[:, 1], c='red', marker='x', s=100, label='Centroids')
        plt.title(f'KMeans Clustering with {best_cluster} clusters')
        plt.xlabel(X.columns[0])
        plt.ylabel(X.columns[1])

        plt.legend()
        plt.grid(True)

        if path:
            os.makedirs(path, exist_ok=True)
            plt.savefig(f"{path}/kmeans_clusters_{best_cluster}.png")
            plt.close()
        else:
            plt.show()
    elif len(X.columns) == 3:
        fig = plt.figure(figsize=(8, 4))
        ax = fig.add_subplot(111, projection='3d')

        ax.scatter(X.iloc[:, 0], X.iloc[:, 1], X.iloc[:, 2], c=kmeans.labels_, cmap='viridis', marker='o')
        ax.scatter(kmeans.cluster_centers_[:, 0], kmeans.cluster_centers_[:, 1], kmeans.cluster_centers_[:, 2], c='red', marker='x', s=100, label='Centroids')
        ax.set_title(f'KMeans Clustering with {best_cluster} clusters')
        ax.set_xlabel(X.columns[0])
        ax.set_ylabel(X.columns[1])
        ax.set_zlabel(X.columns[2])
        plt.legend()
        plt.grid(True)

        if path:
            plt.savefig(f"{path}/kmeans_clusters_{best_cluster}.png")
            plt.close()
        else:
            plt.show()

def get_best_clusters(X, path: str = None):
    """
    Perform clustering on the dataset X and plot silhouette scores for different cluster counts.
    """

    max_silhouette = -1
    silhouette_scores = []
    best_cluster = 0
    for k in range(2, 11):
        kmeans = KMeans(n_clusters=k, random_state=42)
        kmeans.fit(X)
        score = silhouette_score(X, kmeans.labels_)
        silhouette_scores.append(score)
        if score > max_silhouette:
            max_silhouette = score
            best_cluster = k

    # Plot silhouette scores for each k
    plt.figure(figsize=(8, 4))
    plt.plot(range(2, 11), silhouette_scores, marker='o')
    plt.title('Silhouette Scores for Different k')
    plt.xlabel('Number of clusters (k)')
    plt.ylabel('Silhouette Score')
    plt.xticks(range(2, 11))
    plt.grid(True)
    if path:
        os.makedirs(path, exist_ok=True)
        plt.savefig(f"{path}/silhouette_scores.png")
        plt.close()
    else:
        plt.show()
        plt.show()

    # Fit KMeans with the best number of clusters
    kmeans = KMeans(n_clusters=best_cluster, random_state=42)
    kmeans.fit(X)

    radius = _calculate_radius(X, kmeans)

    labels_percentages = np.bincount(kmeans.labels_) / len(kmeans.labels_) * 100

    plot_clusters(X, kmeans, best_cluster, path=path)

    return Cluster_statistics(
        num_clusters=best_cluster,
        silhouette_score=max_silhouette,
        centroids=kmeans.cluster_centers_,
        radius=radius,
        labels_percentages=labels_percentages.tolist()
    )

def get_cluster_defined_number(X, num_clusters: int, path: str = None):
    """
    Perform clustering on the dataset X with a defined number of clusters.
    """

    kmeans = KMeans(n_clusters=num_clusters, random_state=42)
    kmeans.fit(X)
    score = silhouette_score(X, kmeans.labels_)

    radius = _calculate_radius(X, kmeans)

    labels_percentages = np.bincount(kmeans.labels_) / len(kmeans.labels_) * 100

    plot_clusters(X, kmeans, num_clusters, path=path)

    return Cluster_statistics(
        num_clusters=num_clusters,
        silhouette_score=score,
        centroids=kmeans.cluster_centers_,
        radius=radius,
        labels_percentages=labels_percentages
    )

def get_cluster_info_from_json(json_data):
    """
    Extract cluster information from a JSON object.
    """
    num_clusters = json_data.get('num_clusters', 0)
    silhouette_score = json_data.get('silhouette_score', 0.0)
    centroids = [np.array(centroid) for centroid in json_data.get('centroids', [])]
    radius = json_data.get('radius', [])
    labels_percentages = json_data.get('labels_percentages', [])

    return Cluster_statistics(
        num_clusters=num_clusters,
        silhouette_score=silhouette_score,
        centroids=centroids,
        radius=radius,
        labels_percentages=labels_percentages
    )

def compare_clusters(cluster_stats1: Cluster_statistics, cluster_stats2: Cluster_statistics, delta: float = 0.1):
    """
    Compare two cluster statistics objects and visualize the differences.
    """

    return ClusterChanges(original=cluster_stats1, new_data=cluster_stats2, delta=delta)

def clustering_evolution(dfs: list[pd.DataFrame], num_clusters: int):
    """
    Compares the evolution of clustering across multiple DataFrames.
    """

    cluster_stats = []
    for i, df in enumerate(dfs):
        stats = get_cluster_defined_number(df, num_clusters)
        if i!=0:
            stats = ClusterChanges.reorder_changes(cluster_stats[0], stats)
        cluster_stats.append(stats)

    plt.figure(figsize=(10, 6* (1 + num_clusters)))
    plt.suptitle("Evolution of Clustering")

    for i, (title, values, ylabel) in enumerate([
        ('Silhouette Score Evolution', [cs.silhouette_score for cs in cluster_stats], 'Silhouette Score')
    ]):
        plt.subplot(2*(1+num_clusters), 2, i+1)
        plt.plot(values, marker='o', linestyle='-', label=title)
        plt.title(title)
        plt.xlabel('Index')
        plt.ylabel(ylabel)
        plt.grid(True)
        plt.tight_layout(pad=2.0)

    for cluster_idx in range(num_clusters):
        ax = plt.subplot(2*(1+num_clusters), 2, cluster_idx + 3)
        ax.set_title(f"Cluster {cluster_idx + 1}")
        # Gather radius and label percentage values for this cluster across all time steps
        radius_evolution = [cs.radius[cluster_idx] for cs in cluster_stats]
        percentage_evolution = [cs.labels_percentages[cluster_idx] for cs in cluster_stats]
        ax.plot(range(len(cluster_stats)), radius_evolution, marker='o', label='Radius')
        ax.set_xlabel("Index")
        ax.set_ylabel("Radius")
        ax.legend(loc='upper left')
        ax.grid(True)
        ax2 = ax.twinx()
        ax2.plot(range(len(cluster_stats)), percentage_evolution, marker='x', color='red', linestyle='--', label='Label %')
        ax2.set_ylabel("Label Percentage")
        ax2.legend(loc='upper right')

    plt.tight_layout(pad=2.0)
    plt.show()

    # Visualize the evolution of clustering
    for i in range(1, len(cluster_stats)):
        compare_clusters(cluster_stats[i - 1], cluster_stats[i])


import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import os

class DistributionDescriptors:
    """
    A class to represent the distribution descriptors of a single variable.
    """

    def __init__(self, column: pd.Series=None, json_data: dict=None):
        """
        Initializes the DistributionDescriptors from a pandas Series or a JSON representation.
        """
        if column is not None:
            self.mean = column.mean()
            self.std = column.std()
            self.min_val = column.min()
            self.max_val = column.max()
            self.q1 = column.quantile(0.25)
            self.q2 = column.quantile(0.5)
            self.q3 = column.quantile(0.75)
        elif json_data is not None:
            self.mean = json_data['mean']
            self.std = json_data['std']
            self.min_val = json_data['min_val']
            self.max_val = json_data['max_val']
            self.q1 = json_data['q1']
            self.q2 = json_data['q2']
            self.q3 = json_data['q3']
        else:
            raise ValueError("Either a pandas Series or JSON data must be provided to initialize DistributionDescriptors.")        

    def __repr__(self):
        """
        Returns a string representation of the DistributionDescriptors.
        """
        return (f"DistributionDescriptors(mean={self.mean}, std={self.std}, "
                f"min_val={self.min_val}, max_val={self.max_val}, "
                f"q1={self.q1}, q2={self.q2}, q3={self.q3})")
    
    def get_json(self) -> dict:
        """
        Returns a JSON representation of the DistributionDescriptors.
        """
        return {
            "mean": self.mean,
            "std": self.std,
            "min_val": self.min_val,
            "max_val": self.max_val,
            "q1": self.q1,
            "q2": self.q2,
            "q3": self.q3
        }
    
    def __eq__(self, value):
        if not isinstance(value, DistributionDescriptors):
            return NotImplemented

        return (self.mean == value.mean and
                self.std == value.std and
                self.min_val == value.min_val and
                self.max_val == value.max_val and
                self.q1 == value.q1 and
                self.q2 == value.q2 and
                self.q3 == value.q3)

class DistributionChanges:
    """
    A class to represent the changes in distribution between two variables.
    """

    def __init__(self, original: DistributionDescriptors, new_data: DistributionDescriptors, sigma: float = 1.0, delta: float = 0.1):
        """
        Initializes the DistributionChanges with descriptors from two distributions.
        """
        self.original = original
        self.new_data = new_data
        self.sigma = sigma
        self.delta = delta

        self.changed = dict()
        self.unchanged = dict()
        
        metrics = [
            ("mean", original.mean, new_data.mean, sigma * original.std),
            ("std", original.std, new_data.std, delta * original.std),
            ("q1", original.q1, new_data.q1, delta * original.q1),
            ("q2", original.q2, new_data.q2, delta * original.q2),
            ("q3", original.q3, new_data.q3, delta * original.q3),
        ]

        for name, orig_val, new_val, threshold in metrics:
            diff = abs(orig_val - new_val)
            # Avoid division by zero
            if threshold == 0:
                percentage_diff = 0
            else:
                percentage_diff = diff / abs(threshold)
                if diff > abs(threshold):
                    self.changed[name] = int(percentage_diff * 100)
                else:
                    self.unchanged[name] = int(percentage_diff * 100)

    def __repr__(self):
        """
        Returns a string representation of the DistributionChanges.
        """
        change_str = ', '.join([f"{k}: {v}%" for k, v in self.changed.items()])
        unchanged_str = ', '.join([f"{k}: {v}%" for k, v in self.unchanged.items()])

        return f"Changes: {change_str}, Unchanged: {unchanged_str}, Sigma: {self.sigma}, Delta: {self.delta}"

def get_distribution_descriptors(column: pd.Series) -> DistributionDescriptors:
    """
    Returns the distribution descriptors of a given column in a pandas DataFrame.

    Parameters:
    column (pd.Series): The column for which to calculate the distribution descriptors.
    """
    return DistributionDescriptors(column)

def get_distribution_descriptors_from_json(json_data: dict) -> DistributionDescriptors:
    """
    Returns the distribution descriptors from a JSON representation.
    """
    return {column: DistributionDescriptors(json_data=column_data) for column, column_data in json_data.items()}

def get_distribution_descriptors_all_columns(df: pd.DataFrame) -> dict[str, DistributionDescriptors]:
    """
    Returns the distribution descriptors for all columns in a pandas DataFrame.
    """
    return {col: get_distribution_descriptors(df[col]) for col in df.columns}

def _generate_distribution(column: pd.Series):
    """
    Generates a distribution of the data in the column using little balls.
    This function is used internally to plot the distribution of the data.
    """
    _, bins = pd.cut(column, bins=20, retbins=True)
    y_pos = np.zeros(len(column))
    max_pos_bin = np.array([0]*(len(bins)-1))

    for i, data_entry in enumerate(column):
        bin_pos = np.digitize(data_entry, bins)
        # Ensure bin_pos is within valid range
        bin_pos = min(max(bin_pos, 0), len(max_pos_bin) - 1)
        y_pos[i] = max_pos_bin[bin_pos]
        max_pos_bin[bin_pos] += 1

    return y_pos

def plot_distribution_descriptors(column: pd.Series, ax: plt.Axes = None, show: bool = True):
    """
    Plots the distribution descriptors using matplotlib and returns the figure.
    """
    descriptors = get_distribution_descriptors(column)
    if ax is None:
        ax = plt.subplots(figsize=(8, 4))[1]
    y_pos = _generate_distribution(column)

    ax.axvline(descriptors.mean, color='red', linestyle='-', linewidth=2, label=f'Mean: {descriptors.mean:.2f}')
    ax.axvline(descriptors.q1, color='green', linestyle='-', linewidth=2, label=f'Q1: {descriptors.q1:.2f}')
    ax.axvline(descriptors.q2, color='orange', linestyle='-', linewidth=2, label=f'Median (Q2): {descriptors.q2:.2f}')
    ax.axvline(descriptors.q3, color='purple', linestyle='-', linewidth=2, label=f'Q3: {descriptors.q3:.2f}')
    ax.axvline(descriptors.min_val, color='brown', linestyle='-', linewidth=2, label=f'Min: {descriptors.min_val:.2f}')
    ax.axvline(descriptors.max_val, color='pink', linestyle='-', linewidth=2, label=f'Max: {descriptors.max_val:.2f}')
    ax.legend()

    ax.scatter(column, y_pos, s=30, color='blue', alpha=0.6, edgecolors='black')
    ax.set_title(f'Distribution of {column.name}')
    ax.set_xlabel('Value')
    ax.set_ylabel('Frequency (balls)')

    if show:
        plt.show()
    
def plot_distribution_descriptors_all_columns(df: pd.DataFrame, path: str = None):
    """
    Plots the distribution descriptors for all columns in a pandas DataFrame.
    """

    fig, axes = plt.subplots(nrows=len(df.columns), ncols=1, figsize=(10, 5 * len(df.columns)))
    for i, col in enumerate(df.columns):
        plot_distribution_descriptors(df[col], ax=axes[i], show=False)
        axes[i].set_title(f'Distribution of {col}')

    plt.tight_layout()

    if path:
        os.makedirs(path, exist_ok=True)
        plt.savefig(f"{path}/distribution_descriptors_all_columns.png")
        plt.close()
    else:
        plt.show()

def compare_distributions(original: pd.Series, new_data: pd.Series, sigma: float = 1.0, delta: float = 0.1, name: str = None):
    """
    Compares the distributions of two columns in a pandas series.
    """
    changes = DistributionChanges(
        original=get_distribution_descriptors(original),
        new_data=get_distribution_descriptors(new_data),
        sigma=sigma,
        delta=delta
    )

    fig, axes = plt.subplots(1, 2, figsize=(16, 4))
    fig.suptitle(f"Distribution Comparison: {name if name else 'Unnamed'}")
    plot_distribution_descriptors(original, axes[0])
    axes[0].set_title('Original Distribution')

    plot_distribution_descriptors(new_data, axes[1])
    axes[1].set_title('New Data Distribution')

    plt.tight_layout()
    plt.show()

    return changes 

def compare_distribbutions_all_columns(original: pd.DataFrame, new_data: pd.DataFrame, sigma: float = 1.0, delta: float = 0.1):
    """
    Compares the distributions of all columns in two pandas DataFrames.
    """
    for column_name in original.columns:
        original_series = original[column_name]
        new_data_series = new_data[column_name]
        changes = compare_distributions(original_series, new_data_series, sigma=sigma, delta=delta, name=column_name)
        print(f"Changes in column '{column_name}': {changes}")

def descriptor_evolution(dfs: list[pd.Series], name: str = None):
    """
    Compares the evolution of descriptors across.
    """

    descriptors = [get_distribution_descriptors(df) for df in dfs]

    plt.figure(figsize=(10, 6))
    plt.suptitle(f"Evolution of {name}")

    for i, (title, values, ylabel) in enumerate([
        ('Mean Evolution', [d.mean for d in descriptors], 'Mean'),
        ('Standard Deviation Evolution', [d.std for d in descriptors], 'Standard Deviation'),
        ('Q1 Evolution', [d.q1 for d in descriptors], 'Q1'),
        ('Q2 Evolution', [d.q2 for d in descriptors], 'Q2'),
        ('Q3 Evolution', [d.q3 for d in descriptors], 'Q3')
    ], 1):
        plt.subplot(2, 3, i)
        plt.plot(values, marker='o', linestyle='-', label=title)
        plt.title(title)
        plt.xlabel('Index')
        plt.ylabel(ylabel)
        plt.grid(True)
        plt.tight_layout(pad=2.0)

    plt.show()

def descriptor_evolution_all_columns(dfs: list[pd.DataFrame]):
    """
    Compares the evolution of descriptors across different DataFrames.
    """
    for column_name in dfs[0].columns:
        series_list = [df[column_name] for df in dfs]
        descriptor_evolution(series_list, name=column_name)
    



import pandas as pd
from . import univariate as uv
from . import multivariate as mv
import json
import os

def create_initial_report(df: pd.DataFrame, path: str, number_of_output_classes: int = None) -> None:
    """
    Create the initial informative visualizations and statistics for the given DataFrame.
    """
    # Get distribution descriptors for all columns
    descriptors = uv.get_distribution_descriptors_all_columns(df)
    descriptors = {k: v.get_json() for k, v in descriptors.items()}

    os.makedirs(path, exist_ok=True)
    with open(f"{path}/distribution_descriptors.json", 'w+') as f:
        json.dump(descriptors, f, indent=4)

    # Plot distribution descriptors for all columns
    uv.plot_distribution_descriptors_all_columns(df, path=path)

    if number_of_output_classes is not None:
        cluster_info = mv.get_cluster_defined_number(df, number_of_output_classes, path=path)
    else:
        cluster_info = mv.get_best_clusters(df, path=path)

    with open(f"{path}/kmeans_clusters.json", 'w+') as f:
        json.dump(cluster_info.get_json(), f, indent=4)